#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import sys
import logging
import time
from functools import wraps
import numpy as np
from TransportMaps.External import MPI_SUPPORT

__all__ = ['LOG_LEVEL', 'logger', 'deprecate', 'setLogLevel', 'counted',
  	 	   'get_mpi_pool', 'mpi_eval',
           'mpi_map', 'mpi_map_alloc_dmem', 'mpi_alloc_dmem',
           'SumChunkReduce', 'TupleSumChunkReduce',
           'TensorDotReduce', 'ExpectationReduce',
           'AbsExpectationReduce', 'TupleExpectationReduce',
           'distributed_sampling',
		   'generate_total_order_midxs',
           'total_time_cost_function',
           'cached', 'cached_tuple', 'get_sub_cache',
           'taylor_test']

def process_time():
    if sys.version_info >= (3, 3):
        return time.process_time()
    else:
        return time.clock()

####### LOGGING #########
LOG_LEVEL = logging.getLogger().getEffectiveLevel()

logger = logging.getLogger('TM.TransportMaps')
logger.propagate = False
ch = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter("%(asctime)s %(levelname)s:%(name)s: %(message)s",
                              "%Y-%m-%d %H:%M:%S")
ch.setFormatter(formatter)
logger.addHandler(ch)

def deprecate(name, version, msg):
    def deprecate_decorator(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            logger.warning("%s DEPRECATED since v%s. %s" % (name, version, msg))
            return f(*args, **kwargs)
        return wrapped
    return deprecate_decorator

def setLogLevel(level):
    r""" Set the log level for all existing and new objects related to the TransportMaps module

    Args:
      level (int): logging level

    .. see:: the :module:`logging` module.
    """
    import TransportMaps as TM
    TM.LOG_LEVEL = level
    for lname, logger in logging.Logger.manager.loggerDict.items():
        if "TM." in lname:
            logger.setLevel(level)

def counted(f): # Decorator used to count function calls
    @wraps(f)
    def wrapped(slf, *args, **kwargs):
        try:
            x = args[0]
        except IndexError:
            x = kwargs['x']
        try:
            slf.ncalls[f.__name__] += 1
            slf.nevals[f.__name__] += x.shape[0]
        except AttributeError:
            slf.ncalls = {}
            slf.nevals = {}
            slf.ncalls[f.__name__] = 1
            slf.nevals[f.__name__] = x.shape[0]
        except KeyError:
            slf.ncalls[f.__name__] = 1
            slf.nevals[f.__name__] = x.shape[0]
        start = process_time()
        out = f(slf, *args, **kwargs)
        stop = process_time()
        try:
            slf.teval[f.__name__] += (stop-start)
        except AttributeError:
            slf.teval = {}
            slf.teval[f.__name__] = (stop-start)
        except KeyError:
            slf.teval[f.__name__] = (stop-start)
        return out
    return wrapped

            
####### MPI ##########
def get_mpi_pool():
    r""" Get a pool of ``n`` processors
    
    Returns:
      (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`) -- pool of processors

    Usage example::
    
        import numpy as np
        import numpy.random as npr
        from TransportMaps import get_mpi_pool, mpi_map

        class Operator(object):
            def __init__(self, a):
                self.a = a
            def sum(self, x, n=1):
                out = x
                for i in range(n):
                    out += self.a
                return out

        op = Operator(2.)
        x = npr.randn(100,5)
        n = 2

        pool = get_mpi_pool()
        pool.start(3)
        try:
            xsum = mpi_map("sum", op, x, (n,), mpi_pool=pool)
        finally:
            pool.stop()
    """
    if MPI_SUPPORT:
        import mpi_map as mpi_map_mod
        return mpi_map_mod.MPI_Pool_v2()
    else:
        raise RuntimeError("MPI is not supported")

@deprecate('mpi_eval', 'v2.0', 'mpi_map')
def mpi_eval(f, scatter_tuple=None, bcast_tuple=None,
             dmem_key_in_list=None, dmem_arg_in_list=None, dmem_val_in_list=None,
             dmem_key_out_list=None,
             obj=None, reduce_obj=None, reduce_tuple=None, import_set=None,
             mpi_pool=None, splitted=False, concatenate=True):
    r""" Interface for the parallel evaluation of a generic function on points ``x``

    Args:
      f (:class:`object` or :class:`str`): function or string identifying the
        function in object ``obj``
      scatter_tuple (tuple): tuple containing 2 lists of ``[keys]`` and ``[arguments]``
        which will be scattered to the processes.
      bcast_tuple (tuple): tuple containing 2 lists of ``[keys]`` and ``[arguments]``
        which will be broadcasted to the processes.
      dmem_key_in_list (list): list of string containing the keys
        to be fetched (or created with default ``None`` if missing) from the
        distributed memory and provided as input to ``f``.
      dmem_val_in_list (list): list of objects corresponding to the keys defined
        in ``dmem_key_in_list``, used in case we are not executing in parallel
      dmem_key_out_list (list): list of keys to be assigned to the outputs
        beside the first one
      obj (object): object where the function ``f_name`` is defined
      reduce_obj (object): object :class:`ReduceObject` defining the reduce
        method to be applied (if any)   
      reduce_tuple (object): tuple containing 2 lists of ``[keys]`` and ``[arguments]``
        which will be scattered to the processes to be used by ``reduce_obj``
      import_set (set): list of couples ``(module_name,as_field)`` to be imported
        as ``import module_name as as_field``
      mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processors
      splitted (bool): whether the scattering input is already splitted or not
      concatenate (bool): whether to concatenate the output (the output of ``f``
        must be a :class:`numpy.ndarray<numpy.ndarray>` object
    """
    if dmem_key_out_list is None:
        return mpi_map(f=f, scatter_tuple=scatter_tuple, bcast_tuple=bcast_tuple,
                       dmem_key_in_list=dmem_key_in_list,
                       dmem_arg_in_list=dmem_arg_in_list,
                       dmem_val_in_list=dmem_val_in_list,
                       obj=obj, reduce_obj=reduce_obj, reduce_tuple=reduce_tuple,
                       import_set=import_set, mpi_pool=mpi_pool,
                       splitted=splitted, concatenate=concatenate)
    else:
        out = mpi_map_alloc_dmem(
            f=f, scatter_tuple=scatter_tuple, bcast_tuple=bcast_tuple,
            dmem_key_in_list=dmem_key_in_list,
            dmem_arg_in_list=dmem_arg_in_list,
            dmem_val_in_list=dmem_val_in_list,
            dmem_key_out_list=dmem_key_out_list,
            obj=obj, reduce_obj=reduce_obj, reduce_tuple=reduce_tuple,
            import_set=import_set, mpi_pool=mpi_pool,
            splitted=splitted, concatenate=concatenate)
        return (None,) + out
    
def mpi_map(f, scatter_tuple=None, bcast_tuple=None,
            dmem_key_in_list=None, dmem_arg_in_list=None, dmem_val_in_list=None,
            obj=None, obj_val=None,
            reduce_obj=None, reduce_tuple=None, 
            mpi_pool=None, splitted=False, concatenate=True):
    r""" Interface for the parallel evaluation of a generic function on points ``x``

    Args:
      f (:class:`object` or :class:`str`): function or string identifying the
        function in object ``obj``
      scatter_tuple (tuple): tuple containing 2 lists of ``[keys]`` and ``[arguments]``
        which will be scattered to the processes.
      bcast_tuple (tuple): tuple containing 2 lists of ``[keys]`` and ``[arguments]``
        which will be broadcasted to the processes.
      dmem_key_in_list (list): list of string containing the keys
        to be fetched (or created with default ``None`` if missing) from the
        distributed memory and provided as input to ``f``.
      dmem_val_in_list (list): list of objects corresponding to the keys defined
        in ``dmem_key_in_list``, used in case we are not executing in parallel
      obj (object or str): object where the function ``f_name`` is defined
      obj_val (object): object to be used in case not executing in parallel and
        ``obj`` is a string
      reduce_obj (object): object :class:`ReduceObject` defining the reduce
        method to be applied (if any)   
      reduce_tuple (object): tuple containing 2 lists of ``[keys]`` and ``[arguments]``
        which will be scattered to the processes to be used by ``reduce_obj``
      mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processors
      splitted (bool): whether the scattering input is already splitted or not
      concatenate (bool): whether to concatenate the output (the output of ``f``
        must be a :class:`numpy.ndarray<numpy.ndarray>` object
    """
    # Init un-set arguments
    if scatter_tuple is None:
        scatter_tuple = ([],[])
    if bcast_tuple is None:
        bcast_tuple = ([],[])
    if dmem_key_in_list is None:
        dmem_key_in_list = []
    if dmem_arg_in_list is None:
        dmem_arg_in_list = []
    if dmem_val_in_list is None:
        dmem_val_in_list = []
    if reduce_tuple is None:
        reduce_tuple = ([],[])

    # Start evaluation
    if mpi_pool is None:
        # Prepare arguments
        args = {}
        for key, val in zip(scatter_tuple[0], scatter_tuple[1]):
            if splitted:
                if len(val) != 1:
                    raise ValueError("Serial execution but splitted input is %d long" % len(val))
                args[key] = val[0]
            else:
                args[key] = val
        for key, val in zip(bcast_tuple[0], bcast_tuple[1]):
            args[key] = val
        for key, val in zip(dmem_arg_in_list, dmem_val_in_list):
            args[key] = val
        reduce_args = {}
        for key, val in zip(reduce_tuple[0], reduce_tuple[1]):
            if splitted:
                if len(val) != 1:
                    raise ValueError("Serial execution but splitted input is %d long" % len(val))
                reduce_args[key] = val[0]
            else:
                reduce_args[key] = val
        # Retrieve function
        if obj is not None:
            if isinstance(obj, str):
                f = getattr(obj_val, f)
            else:
                f = getattr(obj, f)
        # Evaluate
        fval = f(**args)
        # Reduce if necessary
        if reduce_obj is not None:
            fval = reduce_obj.outer_reduce(
                [ reduce_obj.inner_reduce(fval, **reduce_args) ], **reduce_args )
    else:
        # Prepare arguments
        obj_scatter = mpi_pool.split_data(scatter_tuple[1], scatter_tuple[0],
                                          splitted=splitted)
        obj_bcast = {}
        for key, val in zip(bcast_tuple[0], bcast_tuple[1]):
            obj_bcast[key] = val
        obj_args_reduce = mpi_pool.split_data(reduce_tuple[1], reduce_tuple[0],
                                              splitted=splitted)
        # Evaluate
        fval = mpi_pool.map(f,
                            obj_scatter=obj_scatter, obj_bcast=obj_bcast,
                            dmem_key_in_list=dmem_key_in_list,
                            dmem_arg_in_list=dmem_arg_in_list,
                            obj=obj,
                            reduce_obj=reduce_obj, reduce_args=obj_args_reduce)
        # Put pieces together and return
        if reduce_obj is None and concatenate:
            if isinstance(fval[0], tuple):
                out = []
                for i in range(len(fval[0])):
                    out.append( np.concatenate([fv[i] for fv in fval]) )
                fval = tuple( out )
            else:
                fval = np.concatenate(fval, axis=0)
    return fval
    
def mpi_map_alloc_dmem(f, scatter_tuple=None, bcast_tuple=None,
                       dmem_key_in_list=None, dmem_arg_in_list=None, dmem_val_in_list=None,
                       dmem_key_out_list=None,
                       obj=None, obj_val=None,
                       reduce_obj=None, reduce_tuple=None, 
                       mpi_pool=None, splitted=False, concatenate=True):
    r""" Interface for the parallel evaluation of a generic function on points ``x``

    Args:
      f (:class:`object` or :class:`str`): function or string identifying the
        function in object ``obj``
      scatter_tuple (tuple): tuple containing 2 lists of ``[keys]`` and ``[arguments]``
        which will be scattered to the processes.
      bcast_tuple (tuple): tuple containing 2 lists of ``[keys]`` and ``[arguments]``
        which will be broadcasted to the processes.
      dmem_key_in_list (list): list of string containing the keys
        to be fetched (or created with default ``None`` if missing) from the
        distributed memory and provided as input to ``f``.
      dmem_val_in_list (list): list of objects corresponding to the keys defined
        in ``dmem_key_in_list``, used in case we are not executing in parallel
      dmem_key_out_list (list): list of keys to be assigned to the outputs
        beside the first one
      obj (object): object where the function ``f_name`` is defined
      obj_val (object): object to be used in case not executing in parallel and
        ``obj`` is a string
      reduce_obj (object): object :class:`ReduceObject` defining the reduce
        method to be applied (if any)   
      reduce_tuple (object): tuple containing 2 lists of ``[keys]`` and ``[arguments]``
        which will be scattered to the processes to be used by ``reduce_obj``
      mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processors
      splitted (bool): whether the scattering input is already splitted or not
      concatenate (bool): whether to concatenate the output (the output of ``f``
        must be a :class:`numpy.ndarray<numpy.ndarray>` object
    """
    # Init un-set arguments
    if scatter_tuple is None:
        scatter_tuple = ([],[])
    if bcast_tuple is None:
        bcast_tuple = ([],[])
    if dmem_key_in_list is None:
        dmem_key_in_list = []
    if dmem_arg_in_list is None:
        dmem_arg_in_list = []
    if dmem_val_in_list is None:
        dmem_val_in_list = []
    if dmem_key_out_list is None:
        dmem_key_out_list = []
    if reduce_tuple is None:
        reduce_tuple = ([],[])

    # Start evaluation
    if mpi_pool is None:
        # Prepare arguments
        args = {}
        for key, val in zip(scatter_tuple[0], scatter_tuple[1]):
            if splitted:
                if len(val) != 1:
                    raise ValueError("Serial execution but splitted input is %d long" % len(val))
                args[key] = val[0]
            else:
                args[key] = val
        for key, val in zip(bcast_tuple[0], bcast_tuple[1]):
            args[key] = val
        for key, val in zip(dmem_arg_in_list, dmem_val_in_list):
            args[key] = val
        reduce_args = {}
        for key, val in zip(reduce_tuple[0], reduce_tuple[1]):
            if splitted:
                if len(val) != 1:
                    raise ValueError("Serial execution but splitted input is %d long" % len(val))
                reduce_args[key] = val[0]
            else:
                reduce_args[key] = val
        # Retrieve function
        if obj is not None:
            if isinstance(obj, str):
                f = getattr(obj_val, f)
            else:
                f = getattr(obj, f)
        # Evaluate
        pars = f(**args)
        if not isinstance(pars, tuple):
            pars = (pars,)
    else:
        # Prepare arguments
        obj_scatter = mpi_pool.split_data(scatter_tuple[1], scatter_tuple[0],
                                          splitted=splitted)
        obj_bcast = {}
        for key, val in zip(bcast_tuple[0], bcast_tuple[1]):
            obj_bcast[key] = val
        obj_args_reduce = mpi_pool.split_data(reduce_tuple[1], reduce_tuple[0],
                                              splitted=splitted)
        # Evaluate
        pars = tuple([None] * len(dmem_key_out_list))
        mpi_pool.map_alloc_dmem(
            f, obj_scatter=obj_scatter, obj_bcast=obj_bcast,
            dmem_key_in_list=dmem_key_in_list,
            dmem_arg_in_list=dmem_arg_in_list,
            dmem_key_out_list=dmem_key_out_list,
            obj=obj,
            reduce_obj=reduce_obj, reduce_args=obj_args_reduce)
    return pars

def mpi_alloc_dmem(mpi_pool=None, **kwargs):
    r""" List of keyworded arguments to be allocated in the distributed memory.

    This executes only if an mpi_pool is provided.

    Args:
      mpi_pool (:class:`mpi_map.MPI_Pool<mpi_map.MPI_Pool>`): pool of processors
    """
    if mpi_pool is not None:
        mpi_pool.alloc_dmem(**kwargs)
    
#
# MPI REDUCE OPERATIONS
#
class SumChunkReduce(object):
    r""" Define the summation of the chunks operation.

    The chunks resulting from the output of the MPI evaluation are summed along
    their ``axis``.

    Args:
      axis (tuple [2]): tuple containing list of axes to be used in the
        :func:`sum<numpy.sum>` operation
    """
    def __init__(self, axis=None):
        self.axis = axis
    def inner_reduce(self, x, *args, **kwargs):
        return x
    def outer_reduce(self, x, *args, **kwargs):
        return np.sum(x, axis=self.axis)

class TupleSumChunkReduce(SumChunkReduce):
    r""" Define the summation of the chunks operation over list of tuples.

    The chunks resulting from the output of the MPI evaluation are summed along
    their ``axis``.

    Args:
      axis (tuple [2]): tuple containing list of axes to be used in the
        :func:`sum<numpy.sum>` operation
    """
    def outer_reduce(self, x, *args, **kwargs):
        out = []
        for i in range(len(x[0])):
            xin = [xx[i] for xx in x]
            out.append( super(TupleSumChunkReduce, self).outer_reduce(xin, *args, **kwargs) )
        return tuple(out)

class TensorDotReduce(object):
    r""" Define the reduce tensordot operation carried out through the mpi_map function

    Args:
      axis (tuple [2]): tuple containing list of axes to be used in the
        :func:`tensordot<numpy.tensordot>` operation
    """
    def __init__(self, axis):
        self.axis = axis
    def inner_reduce(self, x, w):
        if x.shape[self.axis[0]] > 0:
            return np.tensordot(x, w, self.axis)
        else:
            return 0.
    def outer_reduce(self, x, w):
        return sum( x )

class ExpectationReduce(TensorDotReduce):
    r""" Define the expectation operation carried out through the mpi_map function
    """
    def __init__(self):
        super(ExpectationReduce, self).__init__((0,0))

class AbsExpectationReduce(ExpectationReduce):
    r""" Define the expectation of the absolute value: :math:`\mathbb{E}[\vert {\bf X} \vert]`
    """
    def inner_reduce(self, x, w):
        return super(AbsExpectationReduce, self).inner_reduce(np.abs(x), w)

class TupleExpectationReduce(ExpectationReduce):
    r""" Define the expectation operation applied on a tuple

    If we are given a tuple :math:`(x_1,x_2)`, the inner reduce
    returns :math:`(\langle x_1,w\rangle , \langle x_2, w\rangle)`.

    Given a list of tuples :math:`\{(x_i,y_i\}_{i=0}^n`, the outer reduce
    gives :math:`(\sum x_i, \sum y_i)`.
    """
    def inner_reduce(self, x, w):
        out = []
        for xx in x:
            out.append( super(TupleExpectationReduce, self).inner_reduce(xx, w) )
        return tuple(out)
    def outer_reduce(self, x, w):
        out = []
        tdim = len(x[0])
        for i in range(tdim):
            xin = [xx[i] for xx in x]
            out.append( super(TupleExpectationReduce, self).outer_reduce(xin, w) )
        return tuple(out)

#
# Distributed operations
#
def distributed_sampling(
        d, qtype, qparams, mpi_pool=None):
    nprocs = mpi_pool.nprocs if mpi_pool is not None else 1
    qparams_split = qparams // nprocs
    qparams_reminder = qparams % nprocs
    qparams_list = [
        qparams_split if i >= qparams_reminder else qparams_split + 1
        for i in range(nprocs) ]
    mass_list = [ float(n)/float(qparams) for n in qparams_list ]
    scatter_tuple = (['qparams', 'mass'],
                     [qparams_list, mass_list])
    bcast_tuple = (['qtype'], [qtype])
    (x, w) = mpi_map_alloc_dmem(
        'quadrature',
        scatter_tuple=scatter_tuple, splitted=True,
        bcast_tuple=bcast_tuple,
        dmem_key_out_list=['x', 'w'],
        obj=d, mpi_pool=mpi_pool,
        concatenate=False)
    return (x, w)
        
#
# Total order multi index generation
#
def generate_total_order_midxs(max_order_list):
    r""" Generate a total order multi-index

    Given the list of maximums :math:`{\bf m}`, the returned set of
    multi-index :math:`I` is such that :math:`\sum_j^d {\bf_i}_j <= max {\bf m}`
    and :math:`{\bf i}_j <= {\bf m}_j`.
    """
    # Improve performances by writing it in cython.
    dim = len(max_order_list)
    max_order = max(max_order_list)
    # Zeros multi-index
    midxs_set = set()
    midx_new = tuple([0]*dim)
    if sum(midx_new) < max_order:
        midxs_old_set = set([ midx_new ]) # Containing increasable multi-indices
    else:
        midxs_old_set = set()
    midxs_set.add(midx_new)
    # Higher order multi-indices
    for i in range(1,max_order+1):
        midxs_new_set = set()
        for midx_old in midxs_old_set:
            for d in range(dim):
                if midx_old[d] < max_order_list[d]:
                    midx_new = list(midx_old)
                    midx_new[d] += 1
                    midxs_set.add( tuple(midx_new) )
                    if sum(midx_new) < max_order:
                        midxs_new_set.add( tuple(midx_new) )
        midxs_old_set = midxs_new_set
    # Transform to list of tuples
    midxs_list = list(midxs_set)
    return midxs_list

#
# Cost functions
#
def total_time_cost_function(
        ncalls, nevals, teval, ncalls_x_solve=None, new_nx=None):
    # Compute elapsed cost as total time
    t = teval.get('log_pdf',0)
    t += teval.get('grad_a_log_pdf',0)
    t += teval.get('tuple_grad_a_log_pdf',0)
    t += teval.get('hess_a_log_pdf',0)
    t += teval.get('action_hess_a_log_pdf',0)
    if new_nx is not None:
        # Compute forecasted time accordingly
        t += ncalls_x_solve.get('log_pdf',0) * new_nx * \
             teval.get('log_pdf',0) / nevals.get('log_pdf',1)
        t += ncalls_x_solve.get('grad_a_log_pdf',0) * new_nx * \
             teval.get('grad_a_log_pdf',0) / nevals.get('grad_a_log_pdf',1)
        t += ncalls_x_solve.get('tuple_grad_a_log_pdf',0) * new_nx * \
             teval.get('tuple_grad_a_log_pdf',0) / nevals.get('tuple_grad_a_log_pdf',1)
        t += ncalls_x_solve.get('hess_a_log_pdf',0) * new_nx * \
             teval.get('hess_a_log_pdf',0) / nevals.get('hess_a_log_pdf',1)
        t += ncalls_x_solve.get('action_hess_a_log_pdf',0) * new_nx * \
             teval.get('action_hess_a_log_pdf',0) / nevals.get('action_hess_a_log_pdf',1)
    return t
    
#
# Caching capabilities (decorator)
#
class cached(object):
    def __init__(self, sub_cache_list=[], caching=True):
        self.sub_cache_list = sub_cache_list
        self.caching = caching
    def __call__(self, f):
        @wraps(f)
        def wrapped(slf, *args, **kwargs):
            try:
                x = args[0]
            except IndexError:
                x = kwargs['x']
            idxs_slice = kwargs.get('idxs_slice', slice(None))
            cache = kwargs.get('cache', None)
            # Decide whether to cache output
            caching = (cache is not None) and self.caching
            # Retrieve from cache
            if caching:
                try:
                    (batch_set, vals) = cache[f.__name__]
                except KeyError as e:
                    new_cache = True
                else:
                    new_cache = False
                    if batch_set[idxs_slice][0]: # Checking only the first
                        return vals[idxs_slice]
            if cache is not None:
                # Init sub-cache if necessary
                for sub_name, sub_len in self.sub_cache_list:
                    try:
                        cache[sub_name + '_cache']
                    except KeyError:
                        if sub_len is None:
                            cache[sub_name + '_cache'] = {'tot_size': cache['tot_size']}
                        elif isinstance(sub_len, int):
                            cache[sub_name + '_cache'] = [
                                {'tot_size': cache['tot_size']}
                                for i in range(sub_len)]
                        elif isinstance(sub_len, str):
                            ll = getattr(slf, sub_len)
                            cache[sub_name + '_cache'] = [
                                {'tot_size': cache['tot_size']}
                                for i in range(ll)]
                        else:
                            raise TypeError("Type of sub_len not recognized")
            # Evaluate function
            out = f(slf, *args, **kwargs)
            # Store in cache
            if caching:
                if new_cache:
                    tot_size = cache['tot_size']
                    cache[f.__name__] = (
                        np.zeros(tot_size, dtype=bool),
                        np.empty((tot_size,)+out.shape[1:], dtype=np.float64))
                    (batch_set, vals) = cache[f.__name__]
                vals[idxs_slice] = out
                batch_set[idxs_slice] = True
            return out
        return wrapped

class cached_tuple(object):
    def __init__(self, commands=[], sub_cache_list=[], caching=True):
        if len(commands) == 0:
            raise AttributeError("You must provide at least one command, " + \
                                 "corresponding to the output on the tuple.")
        self.commands = commands
        self.sub_cache_list = sub_cache_list
        self.caching = caching
    def __call__(self, f):
        @wraps(f)
        def wrapped(slf, *args, **kwargs):
            try:
                x = args[0]
            except IndexError:
                x = kwargs['x']
            idxs_slice = kwargs.get('idxs_slice', slice(None))
            cache = kwargs.get('cache', None)
            # Decide whether to cache output
            caching = (cache is not None) and self.caching
            # Retrieve from cache
            if caching:
                out = [None for i in range(len(self.commands))]
                new_cache = [None for i in range(len(self.commands))]
                out_flag = True
                for i, cmd in enumerate(self.commands):
                    try:
                        (batch_set, vals) = cache[cmd]
                    except KeyError as e:
                        new_cache[i] = True
                        out_flag = False
                    else:
                        new_cache[i] = False
                        if batch_set[idxs_slice][0]: # Checking only the first
                            out[i] = vals[idxs_slice]
                        else:
                            out_flag = False
                if out_flag:
                    return tuple(out)
                else:
                    del out
            if cache is not None:
                # Init sub-cache if necessary
                for sub_name, sub_len in self.sub_cache_list:
                    try:
                        cache[sub_name + '_cache']
                    except KeyError:
                        if sub_len is None:
                            cache[sub_name + '_cache'] = {'tot_size': cache['tot_size']}
                        elif isinstance(sub_len, int):
                            cache[sub_name + '_cache'] = [
                                {'tot_size': cache['tot_size']}
                                for i in range(sub_len)]
                        elif isinstance(sub_len, str):
                            ll = getattr(slf, sub_len)
                            cache[sub_name + '_cache'] = [
                                {'tot_size': cache['tot_size']}
                                for i in range(ll)]
                        else:
                            raise TypeError("Type of sub_len not recognized")
            # Evaluate function
            feval_tuple = f(slf, *args, **kwargs)
            # Store in cache
            if caching:
                for i, (feval, cmd) in enumerate(zip(feval_tuple, self.commands)):
                    if new_cache[i]:
                        tot_size = cache['tot_size']
                        cache[cmd] = (
                            np.zeros(tot_size, dtype=bool),
                            np.empty((tot_size,)+feval.shape[1:], dtype=np.float64))
                    (batch_set, vals) = cache[cmd]
                    vals[idxs_slice] = feval
                    batch_set[idxs_slice] = True
            return feval_tuple
        return wrapped
        
def get_sub_cache(cache, *args):
    out = []
    for arg, dflt in args:
        try:
            out.append( cache[arg + "_cache"] )
        except TypeError:
            out.append( None if dflt is None else [None]*dflt )
    if len(out) > 1:
        return out
    else:
        return out[0]

#
# Taylor test for gradient and Hessian implementations
#
def taylor_test(x, dx, f, gf=None, hf=None, ahf=None, h=1e-4,
                fungrad=False, caching=False,
                args={}):
    r""" Test the gradient and Hessian of a function using the Taylor test.

    Using a Taylor expansion around :math:`{\bf x}`, we have

    .. math::

       f({\bf x}+h \delta{\bf x}) = f({\bf x})
           + h (\nabla f({\bf x}))^\top \delta{\bf x} 
           + \frac{h^2}{2} (\delta{\bf x})^\top \nabla^2 f({\bf x}) \delta{\bf x}
           + \mathcal{O}(h^3)

    Therefore
    
    .. math::

       \vert f({\bf x}+h \delta{\bf x}) - f({\bf x})
       - h (\nabla f({\bf x}))^\top \delta{\bf x} \vert = \mathcal{O}(h^2)

    and
    
    .. math::

       \vert f({\bf x}+h \delta{\bf x}) - f({\bf x})
       - h (\nabla f({\bf x}))^\top \delta{\bf x}
       - \frac{h^2}{2} (\delta{\bf x})^\top \nabla^2 f({\bf x}) \delta{\bf x} \vert
       = \mathcal{O}(h^3)

    Args:
      x (:class:`ndarray<numpy.ndarray>` [:math:`m,d_x`]): evaluation points :math:`{\bf x}`
      dx (:class:`ndarray<numpy.ndarray>` [:math:`m,d_x`]): perturbation direction
        :math:`\delta{\bf x}`
      f (function): function :math:`{\bf x} \mapsto f({\bf x})`. If ``fungrad==True``, then
        ``f`` is the mapping :math:`{\bf x} \mapsto (\nabla f({\bf x}), f({\bf x}))`.
      gf (function): gradient function :math:`{\bf x} \mapsto \nabla f({\bf x})`
      hf (function): Hessian function :math:`{\bf x} \mapsto \nabla^2 f({\bf x})`
      ahf (function): action of the Hessian function
        :math:`{\bf x},\delta{\bf x} \mapsto (\nabla f({\bf x}))^\top \delta{\bf x}`
      h (float): perturbation step
      fungrad (bool): whether ``f`` returns also the gradient or not.
      caching (bool): whether to pass a cache dictionary to the functions.
      args (dict): arguments to be passed to functions
    """
    if caching:
        args['cache'] = {'tot_size': x.shape[0]}
    # Compute at x
    if fungrad:
        fx, gfx = f(x, **args)
    else:
        fx = f(x, **args)
        gfx = gf(x, **args)
    if hf is not None:
        hfx = hf(x, **args)
        ahfx = np.einsum('...ij,...j->...i', hfx, dx)
    elif ahf is not None:
        ahfx = ahf(x, dx, **args)
    # Compute at x + h * dx
    if caching:
        args['cache'] = {'tot_size': x.shape[0]}
    if fungrad:
        fxhdx, _ = f(x + h * dx, **args)
    else:
        fxhdx = f(x + h * dx, **args)
    err_gx1 = np.abs( fxhdx - fx - h * np.einsum('...j,...j->...', gfx, dx) )
    if hf is not None or ahf is not None:
        err_hx1 = np.abs(
            fxhdx - fx - h * np.einsum('...j,...j->...', gfx, dx) - \
            h**2/2 * np.einsum('...i,...i->...', ahfx, dx) )
    # Halve the step
    h /= 2
    if caching:
        args['cache'] = {'tot_size': x.shape[0]}
    if fungrad:
        fxhdx, _ = f(x + h * dx, **args)
    else:
        fxhdx = f(x + h * dx, **args)
    err_gx2 = np.abs( fxhdx - fx - h * np.einsum('...j,...j->...', gfx, dx) )
    if hf is not None or ahf is not None:
        err_hx2 = np.abs(
            fxhdx - fx - h * np.einsum('...j,...j->...', gfx, dx) - \
            h**2/2 * np.einsum('...i,...i->...', ahfx, dx) )
    mrateg = np.min( np.log2(err_gx1/err_gx2) )
    print("Worst convergence rate gradient (should be 2): %.2f" % mrateg)
    if hf is not None or ahf is not None:
        mrateh = np.min( np.log2(err_hx1/err_hx2) )
        print("Worst convergence rate Hessian  (should be 3): %.2f" % mrateh)



'''This file computes the gradient with respect to map coefficients of the
hessian of the log target pdf. It also computes the variance of the generalized
precision matrix omega.'''
import numpy as np

__all__ = ['gen_precision','grad_a_omega','var_omega']

nax = np.newaxis
def gen_precision(pb_dist, data):
    # Compute omega (expectation over samples)
    hess_samps = pb_dist.hess_x_log_pdf(data)
    gen_prec = np.mean(np.abs(hess_samps), axis=0)
    return gen_prec

def grad_a_omega(pb_dist, data):

    grad_omega = pb_dist.grad_a_hess_x_log_pdf(data)
    # note: case where gradient of abs val not defined is not accounted for here
    omega_jac = np.mean(np.multiply(grad_omega, np.sign(grad_omega)), axis=0)

    return omega_jac

def var_omega(pb_dist, data):
    n = data.shape[0]
    dim = pb_dist.transport_map.dim
    omega_CI_jac = grad_a_omega(pb_dist, data)
    fisher_info_samps = pb_dist.hess_a_log_pdf(data)
    fisher_info = - np.mean(fisher_info_samps,axis=0)
    var_a = 1/n * np.linalg.inv(fisher_info)
    var_omega = np.zeros([dim, dim])
    for i in range(dim):
        for j in range(dim):
            temp = np.dot(var_a,omega_CI_jac[:,i,j])
            var_omega[i,j] = np.dot(omega_CI_jac[:,i,j].T,temp)

    return var_omega



'''This file includes multiple ways to reorder the graph nodes, including reverse cholesky, min-fill, and min-degree.'''
import numpy as np
import copy

from sksparse.cholmod import cholesky
from scipy.sparse import csc_matrix
import itertools

__all__ = ['ReverseCholesky', 'MinFill', 'MinDegree']

nax = np.newaxis
class ReverseCholesky():
    def __init__(self):
        self.id = 0

    def perm(self, gp):
        dim = gp.shape[0]
        chol_factor = cholesky(csc_matrix(gp))
        perm = chol_factor.P()
        # reverse ordering
        perm = np.flipud(perm)
        # # inverse perm
        # inv_perm = [0] * dim
        # for i, p in enumerate(perm):
        #     ip[p] = i
        return perm

    def invPerm(self, gp):
        dim = gp.shape[0]
        # inverse perm
        perm = self.perm(gp)
        invPerm = [0] * dim
        for i, p in enumerate(perm):
            invPerm[p] = i
        return invPerm

class MinFill():
    def __init__(self):
        self.id = 1

    def perm(self, gp):
        dim = gp.shape[0]
        gp_copy = copy.copy(gp)
        perm = np.zeros(dim,dtype='int')
        nodes_avail = [i for i in range(dim)]
        for k in range(dim):
            # non_zero_sum = np.zeros(dim);
            added_edges = np.zeros(dim);
            for i in range(dim):
                if i in nodes_avail:
                    # non_zero_sum[i] = np.sum(gp_copy[i,:] != 0)
                    gpLower = np.tril(gp_copy)
                    non_zero_ind  = np.where(gpLower[k,:k] != 0)[0]
                    co_parents = list(itertools.combinations(non_zero_ind,2))
                    for j in range(len(co_parents)):
                        row_index = max(co_parents[j])
                        col_index = min(co_parents[j])
                        if gp_copy[row_index, col_index] == 0:
                            added_edges[i] += 1
                else: 
                    added_edges[i] = dim + 1 # make sure not the minimum
            m, ind = min((s,ind) for ind,s in enumerate(added_edges))
            # set value of permutation vector
            perm[k] = ind
            nodes_avail.remove(ind)
            # if no fill, just set value
            if m == 0:
                pass#print("ok!")
            # else, must fill
            else:
                #marry parents of removed edge
                gpLower = np.tril(gp_copy)
                non_zero_ind  = np.where(gpLower[ind,:ind] != 0)[0]
                co_parents = list(itertools.combinations(non_zero_ind,2))
                for j in range(len(co_parents)):
                    row_index = max(co_parents[j])
                    col_index = min(co_parents[j])
                    gp_copy[row_index, col_index] = 1.0
                    gp_copy[col_index, row_index] = 1.0

        perm = np.flipud(perm)
        perm = np.ndarray.tolist(perm)

        return perm

    def invPerm(self, gp):
        # inverse perm
        perm = self.perm
        invPerm = [0] * dim
        for i, p in enumerate(perm):
            invPerm[p] = i
        return invPerm

class MinDegree():
    def __init__(self):
        self.id = 2

    def perm(self, gp):
        dim = gp.shape[0]
        gp_copy = copy.copy(gp)
        perm = np.zeros(dim,dtype='int')
        nodes_avail = [i for i in range(dim)]
        for k in range(dim):
            non_zero_sum = np.zeros(dim);
            for i in range(dim):
                if i in nodes_avail:
                    non_zero_sum[i] = np.sum(gp_copy[i,:] != 0)
                else: 
                    non_zero_sum[i] = dim + 1 # make sure not the minimum
            m, ind = min((s,ind) for ind,s in enumerate(non_zero_sum))
            #set value of permutation vector
            perm[k] = ind
            #marry parents of removed edge
            gpLower = np.tril(gp_copy)
            non_zero_ind  = np.where(gpLower[k,:k] != 0)[0]
            if len(non_zero_ind) > 1:
                co_parents = list(itertools.combinations(non_zero_ind,2))
                for j in range(len(co_parents)):
                    row_index = max(co_parents[j])
                    col_index = min(co_parents[j])
                    gp_copy[row_index, col_index] = 1.0
                    gp_copy[col_index, row_index] = 1.0
            nodes_avail.remove(ind)

        perm = np.flipud(perm)
        perm = np.ndarray.tolist(perm)

        return perm

    def invPerm(self, gp):
        # inverse perm
        perm = self.perm
        invPerm = [0] * dim
        for i, p in enumerate(perm):
            invPerm[p] = i
        return invPerm



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import logging
import numpy.random as npr

def run_all(log_level=logging.WARNING, ttype='all'):
    # ttype: 'all', 'serial', 'parallel'

    npr.seed(1)

    import sys
    import unittest
    import TransportMaps as TM
    from TransportMaps.tests import test_distributions
    from TransportMaps.tests import test_laplace
    from TransportMaps.tests import test_transportmaps
    from TransportMaps.tests import test_transportmap_distributions
    from TransportMaps.tests import test_transportmap_distributions_sampling
    from TransportMaps.tests import test_kl_divergence
    from TransportMaps.tests import test_kl_minimization
    from TransportMaps.tests import test_L2_misfit
    from TransportMaps.tests import test_L2_minimization
    from TransportMaps.tests import test_xml
    from TransportMaps.tests import test_scripts
    from TransportMaps.tests import test_sequential_inference

    TM.setLogLevel(log_level)

    suites_list = [ test_distributions.build_suite(ttype),
                    test_laplace.build_suite(ttype),
                    test_transportmaps.build_suite(ttype),
                    test_transportmap_distributions.build_suite(ttype),
                    test_transportmap_distributions_sampling.build_suite(ttype),
                    test_kl_divergence.build_suite(ttype),
                    test_kl_minimization.build_suite(ttype),
                    test_L2_misfit.build_suite(ttype),
                    test_L2_minimization.build_suite(ttype),
                    test_xml.build_suite(ttype),
                    test_scripts.build_suite(ttype),
                    test_sequential_inference.build_suite(ttype)]
    all_suites = unittest.TestSuite( suites_list )
    # RUN
    tr = unittest.TextTestRunner(verbosity=2).run(all_suites)
    # Raise error if some tests failed or exited with error state
    nerr = len(tr.errors)
    nfail = len(tr.failures)
    if nerr + nfail > 0:
        print("Errors: %d, Failures: %d" % (nerr, nfail))
        sys.exit(1)

if __name__ == '__main__':
    run_all(ttype='all')



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import logging
import numpy as np
import numpy.linalg as npla

from TransportMaps.Misc import SumChunkReduce, deprecate, mpi_map, mpi_map_alloc_dmem
from TransportMaps.Routines import \
    L2squared_misfit, grad_a_L2squared_misfit, hess_a_L2squared_misfit, \
    storage_hess_a_L2squared_misfit, action_stored_hess_a_L2squared_misfit
from TransportMaps.Functionals.FunctionBase import Function

__all__ = ['ParametricFunctionApproximation',
           'TensorizedFunctionApproximation']

class ParametricFunctionApproximation(Function):
    r""" Abstract class for parametric approximation :math:`f_{\bf a}:\mathbb{R}^d\rightarrow\mathbb{R}` of :math:`f:\mathbb{R}^d\rightarrow\mathbb{R}`.

    Args:
      dim (int): number of dimensions
    """

    def __init__(self, dim):
        super(ParametricFunctionApproximation,self).__init__(dim)
        try:
            self.coeffs
        except AttributeError:
            self.init_coeffs()

    def init_coeffs(self):
        r""" [Abstract] Initialize the coefficients :math:`{\bf a}`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    @property
    def n_coeffs(self):
        r""" [Abstract] Get the number :math:`N` of coefficients :math:`{\bf a}`

        Returns:
          (:class:`int<int>`) -- number of coefficients
        """
        raise NotImplementedError("To be implemented in sub-classes")

    @deprecate("ParametricFunctionApproximation.get_n_coeffs()", "1.0b3",
               "Use property ParametricFunctionApproximation.n_coeffs instead")
    def get_n_coeffs(self):
        return self.n_coeffs

    @property
    def coeffs(self):
        r""" [Abstract] Get the coefficients :math:`{\bf a}`

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`N`]) -- coefficients
        """
        raise NotImplementedError("To be implemented in sub-classes")

    @deprecate("ParametricFunctionApproximation.get_coeffs()", "1.0b3",
               "Use property ParametricFunctionApproximation.coeffs instead")
    def get_coeffs(self):
        return self.coeffs

    @coeffs.setter
    def coeffs(self, coeffs):
        r""" [Abstract] Set the coefficients :math:`{\bf a}`.

        Args:
          coeffs (:class:`ndarray<numpy.ndarray>` [:math:`N`]): coefficients
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def _set_coeffs(self, coeffs):
        self.coeffs = coeffs

    @deprecate("ParametricFunctionApproximation.set_coeffs(value)", "1.0b3",
               "Use setter ParametricFunctionApproximation.coeffs = value instead.")
    def set_coeffs(self, coeffs):
        self.coeffs = coeffs

    def evaluate(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" [Abstract] Evaluate :math:`f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- function evaluations
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def grad_x(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" [Abstract] Evaluate :math:`\nabla_{\bf x} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def grad_a(self, x, precomp=None, idxs_slice=slice(None)):
        r""" [Abstract] Evaluate :math:`\nabla_{\bf a} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            :math:`\nabla_{\bf a} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def hess_a(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" [Abstract] Evaluate :math:`\nabla^2_{\bf a} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N,N`]) --
            :math:`\nabla^2_{\bf a} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def partial_xd(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" [Abstract] Evaluate :math:`\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial_{x_d} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def grad_x_partial_xd(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" [Abstract] Evaluate :math:`\nabla_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def hess_x_partial_xd(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" [Abstract] Evaluate :math:`\nabla^2_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) --
            :math:`\nabla^2_{\bf x}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def partial2_xd(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" [Abstract] Evaluate :math:`\partial^2_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial^2_{x_d} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def grad_a_partial_xd(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" [Abstract] Evaluate :math:`\nabla_{\bf a}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            :math:`\nabla_{\bf a}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def hess_a_partial_xd(self, x, precomp=None, idxs_slice=slice(None), cache=None):
        r""" [Abstract] Evaluate :math:`\nabla^2_{\bf a}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values
          idxs_slice (slice): if precomputed values are present, this parameter
            indicates at which of the points to evaluate. The number of indices
            represented by ``idxs_slice`` must match ``x.shape[0]``.
          cache (:class:`dict`): cache

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,N,N`]) --
            :math:`\nabla^2_{\bf a}\partial_{x_d} f_{\bf a}({\bf x})`
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def precomp_evaluate(self, x, precomp=None):
        r""" [Abstract] Precompute necessary structures for the evaluation of :math:`f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
          (:class:`dict<dict>`) -- data structures
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def precomp_grad_x(self, x, precomp=None):
        r""" [Abstract] Precompute necessary structures for the evaluation of :math:`\nabla_{\bf x} f_{\bf a}` at ``x``

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Return:
          (:class:`dict<dict>`) -- data structures
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def precomp_partial_xd(self, x, precomp=None):
        r""" [Abstract] Precompute necessary structures for the evaluation of :math:`\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
          (:class:`dict<dict>`) -- data structures
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def precomp_grad_x_partial_xd(self, x, precomp=None):
        r""" [Abstract] Precompute  necessary structures for the evaluation of :math:`\nabla_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
          (:class:`dict<dict>`) -- data structures
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def precomp_partial2_xd(self, x, precomp=None):
        r""" [Abstract] Precompute necessary structures for the evaluation of :math:`\partial^2_{x_d} f_{\bf a}` at ``x``.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points

        Returns:
          (:class:`dict<dict>`) -- data structures
        """
        raise NotImplementedError("To be implemented in sub-classes")

    def regression_objective(self, a, params):
        r""" Objective function :math:`\Vert f - f_{\bf a} \Vert^2_{\pi}`

        Args:
          a (:class:`ndarray<numpy.ndarray>` [:math:`N`]): coefficients
          params (dict): dictionary of parameters
        """
        params['nobj'] += 1
        x = params['x']
        w = params['w']
        fvals = params['fvals']
        batch_size = params['batch_size'][0]
        # Update coefficients
        bcast_tuple = (['coeffs'], [a])
        mpi_map("_set_coeffs", bcast_tuple=bcast_tuple,
                obj='f1', obj_val=self,
                mpi_pool=self.mpi_pool, concatenate=False)
        # Evaluate L2 misfit
        scatter_tuple = (['x', 'w', 'f2'], [x, w, fvals])
        bcast_tuple = (['batch_size'], [batch_size])
        dmem_key_in_list = ['f1', 'params1']
        dmem_arg_in_list = ['f1', 'params1']
        dmem_val_in_list = [self, params['params1']]
        reduce_obj = SumChunkReduce(axis=0)
        out = mpi_map(L2squared_misfit, scatter_tuple=scatter_tuple,
                      bcast_tuple=bcast_tuple,
                      dmem_key_in_list=dmem_key_in_list,
                      dmem_arg_in_list=dmem_arg_in_list,
                      dmem_val_in_list=dmem_val_in_list,
                      reduce_obj=reduce_obj,
                      mpi_pool=self.mpi_pool)
        if params['regularization'] is None:
            pass
        elif params['regularization']['type'] == 'L2':
            out += params['regularization']['alpha'] * \
                   npla.norm(a, 2)**2.
        self.logger.debug("Regression Obj. Eval. %d - L2-misfit = %.10e" % (params['nobj'], out))
        return out

    def regression_grad_a_objective(self, a, params):
        r""" Objective function :math:`\nabla_{\bf a} \Vert f - f_{\bf a} \Vert^2_{\pi}`

        Args:
          a (:class:`ndarray<numpy.ndarray>` [:math:`N`]): coefficients
          params (dict): dictionary of parameters
        """
        params['nda_obj'] += 1
        x = params['x']
        w = params['w']
        fvals = params['fvals']
        batch_size = params['batch_size'][1]
        # Update coefficients
        bcast_tuple = (['coeffs'], [a])
        mpi_map("_set_coeffs", bcast_tuple=bcast_tuple,
                obj='f1', obj_val=self,
                mpi_pool=self.mpi_pool, concatenate=False)
        # Evaluate grad_a L2 misfit
        scatter_tuple = (['x', 'w', 'f2'], [x, w, fvals])
        bcast_tuple = (['batch_size'], [batch_size])
        dmem_key_in_list = ['f1', 'params1']
        dmem_arg_in_list = ['f1', 'params1']
        dmem_val_in_list = [self, params['params1']]
        reduce_obj = SumChunkReduce(axis=0)
        out = mpi_map(grad_a_L2squared_misfit,
                      scatter_tuple=scatter_tuple,
                      bcast_tuple=bcast_tuple,
                      dmem_key_in_list=dmem_key_in_list,
                      dmem_arg_in_list=dmem_arg_in_list,
                      dmem_val_in_list=dmem_val_in_list,
                      reduce_obj=reduce_obj,
                      mpi_pool=self.mpi_pool)
        if params['regularization'] is None:
            pass
        elif params['regularization']['type'] == 'L2':
            out += params['regularization']['alpha'] * \
                   2. * (a)
        self.logger.debug("Regression Grad_a Obj. Eval. %d - ||grad_a L2-misfit|| = %.10e" % (
            params['nda_obj'], npla.norm(out)))
        return out

    def regression_hess_a_objective(self, a, params):
        r""" Objective function :math:`\nabla_{\bf a}^2 \Vert f - f_{\bf a} \Vert^2_{\pi}`

        Args:
          a (:class:`ndarray<numpy.ndarray>` [:math:`N`]): coefficients
          params (dict): dictionary of parameters
        """
        params['nda2_obj'] += 1
        x = params['x']
        w = params['w']
        fvals = params['fvals']
        batch_size = params['batch_size'][2]
        # Update coefficients
        bcast_tuple = (['coeffs'], [a])
        mpi_map("_set_coeffs", bcast_tuple=bcast_tuple,
                obj='f1', obj_val=self,
                mpi_pool=self.mpi_pool, concatenate=False)
        # Evaluate hess_a L2 misfit
        scatter_tuple = (['x', 'w', 'f2'], [x, w, fvals])
        bcast_tuple = (['batch_size'], [batch_size])
        dmem_key_in_list = ['f1', 'params1']
        dmem_arg_in_list = ['f1', 'params1']
        dmem_val_in_list = [self, params['params1']]
        reduce_obj = SumChunkReduce(axis=0)
        out = mpi_map(hess_a_L2squared_misfit,
                      scatter_tuple=scatter_tuple,
                      bcast_tuple=bcast_tuple,
                      dmem_key_in_list=dmem_key_in_list,
                      dmem_arg_in_list=dmem_arg_in_list,
                      dmem_val_in_list=dmem_val_in_list,
                      reduce_obj=reduce_obj,
                      mpi_pool=self.mpi_pool)
        if params['regularization'] is None:
            pass
        elif params['regularization']['type'] == 'L2':
            out += np.diag( np.ones(len(self.coeffs))*2.*params['regularization']['alpha'] )
        self.logger.debug("Regression Hess_a Obj. Eval. %d " % params['nda2_obj'])
        return out

    def regression_action_storage_hess_a_objective(self, a, v, params):
        r""" Assemble/fetch Hessian :math:`\nabla_{\bf a}^2 \Vert f - f_{\bf a} \Vert^2_{\pi}` and evaluate its action on :math:`v`

        Args:
          a (:class:`ndarray<numpy.ndarray>` [:math:`N`]): coefficients
          v (:class:`ndarray<numpy.ndarray>` [:math:`N`]): vector on which to apply the Hessian
          params (dict): dictionary of parameters
        """
        x = params['x']
        w = params['w']
        fvals = params['fvals']
        batch_size = params['batch_size'][2]
        # Update coefficients
        bcast_tuple = (['coeffs'], [a])
        mpi_map("_set_coeffs", bcast_tuple=bcast_tuple,
                obj='f1', obj_val=self,
                mpi_pool=self.mpi_pool, concatenate=False)
        # Assemble Hessian
        if not params['hess_assembled']:
            params['nda2_obj'] += 1
            scatter_tuple = (['x', 'w', 'f2'], [x, w, fvals])
            bcast_tuple = (['batch_size'], [batch_size])
            dmem_key_in_list = ['f1', 'params1']
            dmem_arg_in_list = ['f1', 'params1']
            dmem_val_in_list = [self, params['params1']]
            dmem_key_out_list = ['hess_a_L2_misfit']
            (params['hess_a_L2_misfit'], ) = mpi_map_alloc_dmem(
                storage_hess_a_L2squared_misfit, scatter_tuple=scatter_tuple,
                bcast_tuple=bcast_tuple, dmem_key_in_list=dmem_key_in_list,
                dmem_arg_in_list=dmem_arg_in_list, dmem_val_in_list=dmem_val_in_list,
                dmem_key_out_list=dmem_key_out_list, 
                mpi_pool=self.mpi_pool, concatenate=False)
            params['hess_assembled'] = True
            self.logger.debug("Regression Storage Hess_a Obj. Eval. %d " % params['nda2_obj'])
        # Evaluate the action of hess_a L2 misfit
        params['nda2_obj_dot'] += 1
        bcast_tuple = (['v'], [v])
        dmem_key_in_list = ['hess_a_L2_misfit']
        dmem_arg_in_list = ['H']
        dmem_val_in_list = [params['hess_a_L2_misfit']]
        reduce_obj = SumChunkReduce(axis=0)
        out = mpi_map(action_stored_hess_a_L2squared_misfit,
                      bcast_tuple=bcast_tuple,
                      dmem_key_in_list=dmem_key_in_list,
                      dmem_arg_in_list=dmem_arg_in_list,
                      dmem_val_in_list=dmem_val_in_list,
                      reduce_obj=reduce_obj,
                      mpi_pool=self.mpi_pool)
        if params['regularization'] is None:
            pass
        elif params['regularization']['type'] == 'L2':
            regmat = np.diag( np.ones(len(self.coeffs))*2.*params['regularization']['alpha'] )
            out += np.dot(regmat, v)
        self.logger.debug("Regression Action Hess_a Obj. Eval. %d " % params['nda2_obj_dot'])
        return out

    @staticmethod
    def from_xml_element(node, avars, totdim):
        from TransportMaps import XML_NAMESPACE
        import TransportMaps.Functionals as FUNC
        if node.find(XML_NAMESPACE + 'linspan') is not None:
            approx_node = node.find(XML_NAMESPACE + 'linspan')
            return FUNC.LinearSpanApproximation.from_xml_element(approx_node, avars, totdim)
        elif node.find(XML_NAMESPACE + 'monotlinspan') is not None:
            approx_node = node.find(XML_NAMESPACE + 'monotlinspan')
            return FUNC.MonotonicLinearSpanApproximation.from_xml_element(
                approx_node, avars, totdim)
        elif node.find(XML_NAMESPACE + 'intexp') is not None:
            approx_node = node.find(XML_NAMESPACE + 'intexp')
            return FUNC.MonotonicIntegratedExponentialApproximation.from_xml_element(
                approx_node, avars, totdim)
        elif node.find(XML_NAMESPACE + 'intsq') is not None:
            approx_node = node.find(XML_NAMESPACE + 'intsq')
            return FUNC.MonotonicIntegratedSquaredApproximation.from_xml_element(
                approx_node, avars, totdim)
        else:
            raise ValueError("No recognised approximation.")

class TensorizedFunctionApproximation(ParametricFunctionApproximation):
    r""" [Abstract] Class for approximations using tensorization of unidimensional basis

    Args:
      basis_list (list): list of :math:`d` :class:`Basis<SpectralToolbox.Basis>`
      full_basis_list (list): full list of :class:`Basis<SpectralToolbox.Basis>`.
        ``basis_list`` is a subset of ``full_basis_list``. This may be used to
        automatically increase the input dimension of the approximation.
    """
    def __init__(self, basis_list, full_basis_list=None):
        self.basis_list = basis_list
        self.full_basis_list = full_basis_list
        super(TensorizedFunctionApproximation,self).__init__(len(self.basis_list))

    def precomp_evaluate(self, x, precomp=None, precomp_type='uni'):
        r""" Precompute the uni-variate Vandermonde matrices for the evaluation of :math:`f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`list<list>`
            [:math:`d`] of :class:`ndarray<numpy.ndarray>` [:math:`m,n_i`]) --
            dictionary containing the list of Vandermonde matrices
        """
        if precomp is None: precomp = {}
        # Vandermonde matrices
        try: V_list = precomp['V_list']
        except KeyError as e:
            precomp['V_list'] = [ b.GradVandermonde(x[:,i], o)
                                  for i,(b,o) in
                                  enumerate(zip(self.basis_list,
                                                self.get_directional_orders())) ]
        if precomp_type == 'multi':
            self.precomp_Vandermonde_evaluate(x, precomp)
        return precomp

    def precomp_grad_x(self, x, precomp=None):
        r""" Precompute the uni-variate Vandermonde matrices for the evaluation of :math:`\nabla_{\bf x} f_{\bf a}` at ``x``

        Letting :math:`\Phi^{(i)}(x_i)` being the uni-variate Vandermonde in
        :math:`x_i`, the ``i``-th element of the returned list is
        :math:`\partial_{x_i}\Phi^{(i)}(x_i)`.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Return:
          (:class:`dict<dict>` with :class:`list<list>`
            [:math:`d`] of :class:`ndarray<numpy.ndarray>` [:math:`m,n_i`]) --
            dictionary containing the list of Vandermonde matrices
        """
        if precomp is None: precomp = {}
        try: V_list = precomp['V_list']
        except KeyError as e:
            self.precomp_evaluate(x, precomp)
        try: partial_x_V_list = precomp['partial_x_V_list']
        except KeyError as e:
            partial_x_V_list = [ b.GradVandermonde(x[:,i], o, k=1)
                                 for i,(b,o)
                                 in enumerate(zip(self.basis_list,
                                                  self.get_directional_orders())) ]
            precomp['partial_x_V_list'] = partial_x_V_list
        return precomp

    def precomp_hess_x(self, x, precomp=None):
        r""" Precompute the uni-variate Vandermonde matrices for the evaluation of :math:`\nabla^2_{\bf x} f_{\bf a}` at ``x``

        Letting :math:`\Phi^{(i)}(x_i)` being the uni-variate Vandermonde in
        :math:`x_i`, the ``i``-th element of the returned list is
        :math:`\partial^2_{x_i}\Phi^{(i)}(x_i)`.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Return:
          (:class:`dict<dict>` with :class:`list<list>`
            [:math:`d`] of :class:`ndarray<numpy.ndarray>` [:math:`m,n_i`]) --
            dictionary containing the list of Vandermonde matrices
        """
        if precomp is None: precomp = {}
        try: V_list = precomp['V_list']
        except KeyError as e:
            self.precomp_evaluate(x, precomp)
        try: partial_x_V_list = precomp['partial_x_V_list']
        except KeyError as e:
            self.precomp_grad_x(x, precomp)
        try: partial2_x_V_list = precomp['partial2_x_V_list']
        except KeyError as e:
            partial2_x_V_list = [ b.GradVandermonde(x[:,i], o, k=2)
                                  for i,(b,o) in enumerate(zip(
                                          self.basis_list, self.get_directional_orders())) ]
            precomp['partial2_x_V_list'] = partial2_x_V_list
        return precomp

    def precomp_partial_xd(self, x, precomp=None, precomp_type='uni'):
        r""" Precompute uni-variate Vandermonde matrix for the evaluation of :math:`\partial_{x_d} f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`ndarray<numpy.ndarray>` [:math:`m,n_d`]) --
            dictionary with Vandermonde matrix
        """
        if precomp is None: precomp = {}
        try: V_list = precomp['V_list']
        except KeyError as e:
            self.precomp_evaluate(x, precomp)
        try: partial_xd_V_last = precomp['partial_xd_V_last']
        except KeyError as e:
            o = self.get_directional_orders()[-1]
            precomp['partial_xd_V_last'] = self.basis_list[-1].GradVandermonde(
                x[:,-1], o, k=1)
        if precomp_type == 'multi':
            self.precomp_Vandermonde_partial_xd(x, precomp)
        return precomp

    def precomp_partial2_xd(self, x, precomp=None):
        r""" Precompute uni-variate Vandermonde matrix for the evaluation of :math:`\partial^2_{x_d} f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`ndarray<numpy.ndarray>` [:math:`m,n_d`]) --
            dictionary with Vandermonde matrix
        """
        if precomp is None: precomp = {}
        try: V_list = precomp['V_list']
        except KeyError as e:
            self.precomp_evaluate(x, precomp)
        try: partial2_xd_V_last = precomp['partial2_xd_V_last']
        except KeyError as e:
            o = self.get_directional_orders()[-1]
            precomp['partial2_xd_V_last'] = self.basis_list[-1].GradVandermonde(x[:,-1], o, k=2)
        return precomp

    def precomp_grad_x_partial_xd(self, x, precomp=None):
        r""" Precompute uni-variate Vandermonde matrices for the evaluation of :math:`\nabla_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`list<list>` [d]
            :class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            dictionary containing the list of uni-variate Vandermonde matrices.
        """
        if precomp is None: precomp = {}
        try: V_list = precomp['V_list']
        except KeyError as e:
            self.precomp_evaluate(x, precomp)
        try: partial_x_V_list = precomp['partial_x_V_list']
        except KeyError as e:
            self.precomp_grad_x(x, precomp)
        try: partial2_xd_V_last = precomp['partial2_xd_V_last']
        except KeyError as e:
            self.precomp_partial2_xd(x, precomp)
        return precomp

    def precomp_partial3_xd(self, x, precomp=None):
        r""" Precompute uni-variate Vandermonde matrix for the evaluation of :math:`\partial^3_{x_d} f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`ndarray<numpy.ndarray>` [:math:`m,n_d`]) --
            dictionary with Vandermonde matrix
        """
        if precomp is None: precomp = {}
        try: V_list = precomp['V_list']
        except KeyError as e:
            self.precomp_evaluate(x, precomp)
        try: partial3_xd_V_last = precomp['partial3_xd_V_last']
        except KeyError as e:
            o = self.get_directional_orders()[-1]
            precomp['partial3_xd_V_last'] = self.basis_list[-1].GradVandermonde(x[:,-1], o, k=3)
        return precomp

    def precomp_hess_x_partial_xd(self, x, precomp=None):
        r""" Precompute uni-variate Vandermonde matrices for the evaluation of :math:`\nabla_{\bf x}\partial_{x_d} f_{\bf a}` at ``x``.

        Enriches the ``precomp`` dictionary if necessary.

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points
          precomp (:class:`dict`): dictionary of precomputed values

        Returns:
          (:class:`dict<dict>` with :class:`list<list>` [d]
            :class:`ndarray<numpy.ndarray>` [:math:`m,N`]) --
            dictionary containing the list of uni-variate Vandermonde matrices.
        """
        if precomp is None: precomp = {}
        try: V_list = precomp['V_list']
        except KeyError as e:
            self.precomp_evaluate(x, precomp)
        try: partial_x_V_list = precomp['partial_x_V_list']
        except KeyError as e:
            self.precomp_grad_x(x, precomp)
        try: partial2_x_V_list = precomp['partial2_x_V_list']
        except KeyError as e:
            self.precomp_hess_x(x, precomp)
        try: partial3_xd_V_last = precomp['partial3_xd_V_last']
        except KeyError as e:
            self.precomp_partial3_xd(x, precomp)
        return precomp


#!/usr/bin/env python

#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import logging
import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as stats
from tabulate import tabulate

import TransportMaps as TM
import TransportMaps.Distributions as DIST
import TransportMaps.Maps as MAPS

import xun   # The Python wrapper to the C++ sampler/samples

# logging.basicConfig(level=logging.INFO)

PLOTTING = True
ORDER = 3
TOL = 1e-5

# REGULARIZATION
# None: No regularization
# L2: L2 regularization
REG = None
# REG = {'type': 'L2',
#        'alpha': 1e-3}

# L2 error estimation
# 0: Monte Carlo quadrature with n point (the samples)
qtype = 0

# Gradient information
# 0: derivative free
# 1: gradient information
# 2: Hessian information
ders = 2

# MC samples
N_samp = 1000

# Define target distribution pi (the one we have samples for)
# We just need to define the rvs and quadrature functions, which will
# both return random samples.
# For further documentation check out the class TransportMap.Densities.Distribution
class WeirdDistribution(DIST.Distribution):
    def __init__(self):
        self.dim = 1
    def rvs(self, n):
        return xun.weird_samples(n)
    def quadrature(self, qtype, qparams):
        if qtype == 0:
            x = self.rvs(qparams)
            w = np.ones(qparams)/float(qparams)
        else: raise ValueError("Quadrature type not recognized")
        return (x,w)
    # For the interest of the example, we actually know the pdf and log_pdf
    # but they are in general not needed in the construction from samples
    def pdf(self, x, params=None):
        a = 2.
        b = 4.
        return a/b * (x/b)**(a-1.) * np.exp(-(x/b)**a)
    # def log_pdf(self, x, params=None): # We really don't need this here.
    #     a = 2.
    #     b = 4.
    #     return np.log(a) - np.log(b) + (a-1.) * (np.log(x)-np.log(b)) - (x/b)**a

pi = WeirdDistribution()

# Generate samples from the distribution and visualize
tar_samp = pi.rvs(N_samp)
xx = np.linspace(np.min(tar_samp), np.max(tar_samp), 100)
plt.figure()
plt.hist(tar_samp, bins=20, normed=True, histtype='step')
plt.plot(xx, pi.pdf(xx))
plt.show(False)

# Define the base distribution \rho (standard normal)
rho = DIST.StandardNormalDistribution(pi.dim)

# Define the support map L, i.e. the object that maps samples in the
# real line (reference support) to samples in the half-line (target support).
# For further documentation check out the file
# TransportMaps/Maps/FrozenTriangularTransportMaps.py
support_map = MAPS.FrozenExponentialDiagonalTransportMap(1)

# Plot the generated samples on the real line
supp_samp = support_map.inverse( tar_samp )
supp_xx = np.linspace(np.min(supp_samp), np.max(supp_samp), 100)
plt.figure()
plt.hist(supp_samp, bins=20, normed=True, histtype='step')
plt.show(False)

# Build the transport map (isotropic for each entry)
# See documentation of the constructor for more information.
active_vars = [ [0] ]
tm_approx = TM.Default_IsotropicIntegratedExponentialTriangularTransportMap(
    1, ORDER, 'total', active_vars=active_vars)
print("Number of coefficients: %d" % tm_approx.get_n_coeffs())

# Set up the arguments of D_KL(T_\sharp L^\sharp \pi || \rho)
L_pull_pi = DIST.PullBackTransportMapDistribution(support_map, pi)
T_push_L_pull_pi = DIST.PushForwardTransportMapDistribution(tm_approx, L_pull_pi)

# SOLVE
log_entry_solve = T_push_L_pull_pi.minimize_kl_divergence(
    rho, qtype=qtype, qparams=N_samp, regularization=REG,
    tol=TOL, ders=ders, batch_size=[None,None,None])

# Construct approximate distribution T^\sharp \pi
T_push_rho = DIST.PullBackTransportMapDistribution(tm_approx, rho)
# Construct approximate distribution L_\sharp T^\sharp \pi
L_pull_T_push_rho = DIST.PushForwardTransportMapDistribution(
    support_map, T_push_rho)

# Visualize approximated pdf
plt.figure()
plt.plot(xx, pi.pdf(xx), label='Exact')
plt.plot(xx, L_pull_T_push_rho.pdf(xx[:,np.newaxis]), label='Approximation')
plt.legend()
plt.show(False)

# Sample from the approximated distribution
app_samp = L_pull_T_push_rho.rvs(N_samp)
plt.figure()
plt.plot(xx, pi.pdf(xx), label='Exact')
plt.hist(app_samp, bins=20, normed=True, histtype='step')
plt.legend()
plt.show(False)


# # Push forward reference samples in two steps
# # (1) Map x from \pi to y from T^\sharp \pi
# # (2) Map y from T^\sharp \pi to z from L_\sharp T^\sharp \pi
# tmp = approx_distribution_1.map_samples_base_to_target( ref_samp )
# fapp_samp = approx_distribution.map_samples_base_to_target( tmp ) 

# # PLOT
# if PLOTTING:
#     # Map to T^\sharp \pi
#     t1_approx_tmp = approx_distribution_1.map_samples_base_to_target(X1d)
#     # Map to L_\sharp T^\sharp \pi
#     t1_approx = approx_distribution.map_samples_base_to_target(t1_approx_tmp)[:,0]
#     # Plot 1d transport map
#     ax_map_1d.plot(X1d[:,0], t1_approx, label='ord %d' % order)

#     # Plot the 1d distribution
#     if setup['dim'] == 1:
#         pdf1d_approx = approx_distribution.pdf(X1d_kde)
#     else:
#         t1_kde_approx = stats.gaussian_kde(fapp_samp[:,0][:,np.newaxis].T)
#         pdf1d_approx = t1_kde_approx(X1d_kde.T)
#     ax_kde_1d.plot(X1d_kde, pdf1d_approx, label='ord %d' % order)

#     if setup['dim'] >= 2:
#         # Map to T_\sharp \pi
#         t2_approx_tmp = approx_distribution_1.map_samples_base_to_target(X2d)
#         # Map to L_\sharp T_\sharp \pi
#         t2_approx = approx_distribution.map_samples_base_to_target(t2_approx_tmp)[:,1]
#         # Plot 2d transport map
#         ax_maps[i_ord].contour(xx2d, yy2d, t2_approx.reshape(xx2d.shape),
#                                linestyles='dashed', levels=levels_maps2d)

#         # Plot the 2d distribution
#         if setup['dim'] == 2:
#             pdf2d_approx = approx_distribution.pdf(X2d_kde)
#         else:
#             t2_kde_approx = stats.gaussian_kde(fapp_samp[:,:1].T)
#             pdf2d_approx = t2_kde_approx(X2d_kde.T)
#         axs_pdf2d[i_ord].contour(xx2d,yy2d,pdf2d_approx.reshape(xx2d.shape),
#                                  linestyles='dashed', levels=levels_pdf2d)
#         axs_pdf2d[i_ord].scatter(fapp_samp[:nscatter,0],fapp_samp[:nscatter,1],color='r')

# # Compute L2 error
# dist = np.sqrt( np.sum((T_samp - fapp_samp)**2., axis=1) )
# log_entry.append( np.sqrt( np.sum( dist**2. ) / float(N_samp) ) )

# log.append( log_entry )

# if PLOTTING:
#     fig_map_1d.gca().legend()
#     fig_kde_1d.gca().legend()
#     plt.show(False)

# # # Decorate map plotting
# # # ax_map.set_ylim([f(X[0]), f(X[-1])])
# # ax_map.scatter(x, np.zeros(x.shape))
# # ax_map.set_xlabel('x')
# # ax_map.set_ylabel('$T(x)$')
# # # ax_map.set_ylabel('$ (F^{-1} \circ F_{ref})x $')
# # # ax_map.set_ylim([np.min(exact_tm),np.max(exact_tm)])
# # ax_map.legend(loc='best')
# # ax_map.grid(True)

# # # fig_map.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-Approximation.pdf' % \
# # #                 (title, tit_type, tit_reg, tit_intest))
# # # fig_map.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-Approximation.eps' % \
# # #                 (title, tit_type, tit_reg, tit_intest))

# # # Decorate kde plotting
# # ax_kde.set_xlabel('x')
# # ax_kde.set_ylabel('PDF')
# # ax_kde.legend(loc='best')

# # # fig_kde.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-DistributionApproximation.pdf' % \
# # #                 (title, tit_type, tit_reg, tit_intest))
# # # fig_kde.savefig('Figs/KLdiv-%s-%s-%s-%s-MonotoneMap1d-DistributionApproximation.eps' % \
# # #                 (title, tit_type, tit_reg, tit_intest))

# # plt.show(False)

# print(" ")
# print(tabulate(log, headers=log_header))
# print(" ")

# print(tabulate(log, headers=log_header, tablefmt="latex", floatfmt=".2e"))


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import unittest
import numpy as np
import numpy.random as npr

from TransportMaps.RandomizedLinearAlgebra import \
    adaptive_randomized_range_finder, randomized_direct_eigenvalue_decomposition

class RLATest(unittest.TestCase):

    def test_adaptive_randomized_range_finder(self):
        M = 50
        N = 70
        k = 5
        r = 10
        eps = 1e-6
        A1 = npr.randn(M, k)
        A2 = npr.randn(N, k)
        A = np.dot(A1, A2.T)
        def action(X, A):
            return np.dot(A, X)
        def action_transpose(X, A):
            return np.dot(A.T, X)
        power_n = 1
        kwargs = {'A': A}
        Q = adaptive_randomized_range_finder(
            action, N, M, eps, r, kwargs=kwargs,
            power_n=power_n, action_transpose=action_transpose)
        self.assertTrue(Q.shape[1] == r+k-1)
        B = action_transpose(Q, A)
        C = Q
        A1 = np.dot(C, B.T)
        self.assertTrue(np.allclose(A, A1))

    def test_randomized_direct_eigenvalue_decomposition(self):
        N = 70
        k = 5
        r = 10
        eps = 1e-6
        A = npr.randn(N, k)
        A = np.dot(A, A.T)
        def action(X, A):
            return np.dot(A, X)
        def action_transpose(X, A):
            return np.dot(A.T, X)
        power_n = 1
        kwargs = {'A': A}
        D, U = randomized_direct_eigenvalue_decomposition(
            action, N, eps, r, kwargs=kwargs,
            power_n=power_n)
        A1 = np.dot(U, (D[np.newaxis,:] * U).T)
        self.assertTrue(np.allclose(A, A1))
        
def build_suite(*args, **kwargs):
    suite_rla = unittest.TestLoader().loadTestsFromTestCase( RLATest )
    # Group suites
    suites_list = [suite_rla]
    all_suites = unittest.TestSuite( suites_list )
    return all_suites

def run_tests(*args, **kwargs):
    all_suites = build_suite()
    # RUN
    unittest.TextTestRunner(verbosity=2).run(all_suites)

if __name__ == '__main__':
    run_tests()
